# Generated by Django 4.2 on 2023-05-10 11:08

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("permissions", "0004_create_anonymous_user"),
        ("observers", "0003_alter_observer_object_id"),
    ]

    operations = [
        migrations.AlterField(
            model_name="backgroundtask",
            name="permission_group",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="%(class)s",
                to="permissions.permissiongroup",
            ),
        ),
    ]
