# Generated by Django 3.2.16 on 2023-03-31 08:44

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("observers", "0002_backgroundtask"),
    ]

    operations = [
        migrations.AlterField(
            model_name="observer",
            name="object_id",
            field=models.PositiveIntegerField(default=0),
            preserve_default=False,
        ),
    ]
